import typing as t
from abc import ABC
from abc import abstractmethod

from acb.config import Config
from acb.config import Settings
from acb.depends import depends


class SecretsBaseSettings(Settings):
    ...


class SecretsBase(ABC):
    config: Config = depends()

    @abstractmethod
    async def init(self) -> t.NoReturn:
        raise NotImplementedError

    @abstractmethod
    async def list(self, adapter: str) -> t.NoReturn:
        raise NotImplementedError

    @abstractmethod
    async def create(self, name: str, value: str) -> t.NoReturn:
        raise NotImplementedError

    @abstractmethod
    async def update(self, name: str, value: str) -> t.NoReturn:
        raise NotImplementedError

    @abstractmethod
    async def get(self, name: str) -> t.NoReturn:
        raise NotImplementedError

    @abstractmethod
    async def delete(self, name: str) -> t.NoReturn:
        raise NotImplementedError



import typing as t
from calendar import isleap
from calendar import monthrange
from datetime import date
from datetime import datetime
from datetime import timedelta
from itertools import chain
from pathlib import Path
from re import search

from sqlmodel import select
from sqlmodel import SQLModel

import arrow
from acb.actions.encode import load
from acb.adapters.logger import Logger
from acb.adapters.sql import Sql
from acb.adapters.storage import Storage
from acb.config import Config
from acb.depends import depends
from aiopath import AsyncPath
from pydantic import BaseModel
from pydantic import create_model
from ._model import SqlModel


# from contextlib import suppress
# from sqlalchemy.exc import IntegrityError
# from sqlalchemy.exc import InvalidRequestError
# from sqlalchemy.orm.exc import UnmappedInstanceError


class SqlBackupUtils(BaseModel):
    storage: Storage = depends()  # type: ignore
    config: Config = depends()  # type: ignore
    backup_path: t.Optional[Path] = None

    @staticmethod
    def get_timestamp(name: str) -> str | None:
        pattern = r"(.*)(-)(?P<timestamp>[\d]{10})(-)(.*)"
        match = search(pattern, name)
        return match.group("timestamp") or None

    def get_files(self) -> t.Generator[t.Any, t.Any, t.Any]:
        blobs = self.storage.sql.list()
        for blob in blobs:
            if self.get_timestamp(blob.name):
                yield blob.name

    def get_timestamps(self) -> list[int]:
        if not self.files:
            self.files = tuple(self.get_files())
        different_timestamps = []
        for name in self.files:
            timestamp = self.get_timestamp(name)
            if timestamp and timestamp not in different_timestamps:
                different_timestamps.append(int(timestamp))
        return different_timestamps

    def by_timestamp(self, timestamp: str) -> t.Generator[t.Any, t.Any, t.Any]:
        if not self.files:
            self.files = tuple(self.get_files())
        for name in self.files:
            if timestamp == self.get_timestamp(name):
                yield name

    def valid(self, timestamp: int) -> bool:
        if timestamp in self.get_timestamps():
            return True
        # print('==> Invalid id. Use "history" to list existing downloads')
        return False

    def get_path(
        self,
        class_name: str = "",
        timestamp: t.Any = arrow.utcnow().timestamp,
    ) -> Path:
        self.backup_path = Path(
            f"{self.config.app.name}-{int(timestamp)}-{class_name}.json"
        )
        return self.backup_path


class SqlBackupDates(BaseModel, arbitrary_types_allowed=True):
    today: arrow.Arrow = arrow.utcnow()
    white_list: list[int] = []
    black_list: list[int] = []
    dates: list[int] = []

    def __init__(self, dates: t.Optional[list[int]] = None, **data: t.Any) -> None:
        super().__init__(**data)
        self.dates = sorted(dates, reverse=True) if dates else []
        self.process()  # feed self.white_list & self.black_list

    def get_last_month_length(self) -> int:
        return monthrange(self.today.year, self.today.shift(months=-1).month)[1]

    def get_last_year_length(self) -> int:
        first_day = date(self.today.year, 1, 1)  # current year
        last_day = first_day - timedelta(days=1)  # last year
        return 366 if isleap(last_day.year) else 365

    def filter_dates(
        self, dates: list[t.Any], period: str = "week"
    ) -> t.Generator[t.Any, t.Any, t.Any]:
        reference = self.today.int_timestamp
        method_mapping: dict[str, t.Any] = {
            "week": lambda obj: getattr(obj, "isocalendar")()[1],  # type: ignore
            "month": lambda obj: getattr(obj, "month"),  # type: ignore
            "year": lambda obj: getattr(obj, "year"),  # type: ignore
        }
        for dt in dates:
            comp_date = datetime.fromtimestamp(int(dt))
            comparison = method_mapping.get(period)(comp_date)
            ref_date = datetime.fromtimestamp(reference)
            if comparison != method_mapping.get(period)(ref_date):
                reference = dt
                yield dt

    def process(self) -> None:
        last_w = self.today.shift(days=-7).int_timestamp
        last_m = self.today.shift(days=-(self.get_last_month_length())).int_timestamp
        last_y = self.today.shift(days=-(self.get_last_year_length())).int_timestamp
        backups_week = []
        backups_month = []
        backups_year = []
        backups_older = []
        for timestamp in self.dates:
            if timestamp >= last_w:
                backups_week.append(timestamp)
            elif timestamp >= last_m:
                backups_month.append(timestamp)
            elif timestamp >= last_y:
                backups_year.append(timestamp)
            else:
                backups_older.append(timestamp)
        self.white_list.extend(
            chain(
                backups_week,
                self.filter_dates(backups_month),
                self.filter_dates(backups_year, "month"),
                self.filter_dates(backups_older, "year"),
            )
        )
        diff_as_list = [d for d in self.dates if d not in self.white_list]
        self.black_list.extend(sorted(diff_as_list, reverse=True))


class SqlBackup(SqlBackupDates, SqlBackupUtils):
    sql: Sql = depends()  # type: ignore
    logger: Logger = depends()  # type: ignore
    do_not_backup: list[str] = []
    models: list[t.Any] = []
    sure_delete: bool = False

    @staticmethod
    def show() -> list[t.Any]:
        return [
            m
            for m in SQLModel.metadata.schema  # type: ignore
            if isinstance(m, type) and issubclass(m, SqlModel)
        ]

    def get_mapped_classes(self) -> list[t.Any]:
        self.add_subclasses(SqlModel)
        return self.models

    def add_subclasses(self, model: t.Any) -> None:
        if model.__subclasses__():
            for submodel in model.__subclasses__():
                self.add_subclasses(submodel)
        else:
            self.models.append(model)

    async def get_data(self) -> dict[str, t.Any]:
        data = {}
        async with self.sql.session() as session:
            for model in self.get_mapped_classes():
                query = select(model)
                results = session.exec(query)
                data[model.__name__] = results.model_dump_to_json()
        return data

    async def backup(self, class_name: str, data: t.Any, timestamp: int) -> t.NoReturn:
        path = self.get_path(class_name=class_name, timestamp=timestamp)
        await self.storage.sql.save(data, path)

    async def save(self) -> t.NoReturn:
        data = await self.get_data()
        timestamp = arrow.utcnow().int_timestamp

        for class_name in data.keys():
            path = self.get_path(class_name=class_name, timestamp=timestamp)
            self.logger.debug(f"Backing up - {path.name}")
            await self.backup(
                class_name=class_name, data=data[class_name], timestamp=timestamp
            )

    async def get_backups(self) -> list[int]:
        timestamps = self.get_timestamps()
        backups = {}
        for timestamp in timestamps:
            adate = arrow.Arrow.fromtimestamp(timestamp)
            date_formatted = adate.humanize()
            backups[timestamp] = []
            fps = [
                AsyncPath(b)
                async for b in self.storage.sql.list()
                if timestamp in b.name
            ]
            for fp in fps:
                if await self.storage.sql.exists(fp):
                    backups[timestamp].append(fp.name)
            adate_formatted = adate.format("MM-DD-YYYY HH:mm:ss ZZ")
            self.logger.debug(f"{date_formatted} ==> {adate_formatted}:")
        return sorted(backups, reverse=True)

    def get_last_backup(self) -> bool | int:
        timestamps = self.get_timestamps()
        if not timestamps:
            return False
        timestamp = sorted(timestamps, reverse=True)[0]
        adate = arrow.Arrow.fromtimestamp(timestamp)
        self.logger.info(f"Last backup: {adate.format('MM-DD-YYYY HH:mm:ss ZZ')}")
        return timestamp

    async def restore_rows(self, path: AsyncPath) -> t.Any:
        contents = await self.storage.sql.read(path)
        pattern = r"-(\D+).json"
        model = search(pattern, path.name).group(1)
        if contents:
            loaded_model = create_model(
                model,
                __base__=SqlModel,
                __cls_kwargs__={"table": True},
                **load.json(contents),
            ).save()
            if loaded_model:
                async with self.sql.session() as session:
                    for m in self.show():
                        m_name = m.__name__
                        if m_name == model:
                            await session.query(m).delete()
                    await session.commit()

    async def restore_backup(self, timestamp: int) -> None:
        paths = [
            AsyncPath(b.name)
            async for b in self.storage.sql.list()
            if (timestamp in b.name)
        ]
        when = arrow.Arrow.fromtimestamp(timestamp).humanize()
        self.logger.info(f"Restoring backup from {when}.....")
        for path in paths:
            self.logger.debug(f"\tRestoring:  {path.name}")
            await self.restore_rows(path)
            # status, fails = await self.restore_rows(path)
            # self.logger.info(status)
            # for f in fails:
            #     self.logger.error(f"\t\tRestore of {f} failed!")

    def delete_backups(self, delete_list: list[str]) -> None:
        delete_me = self.storage.sql.delete(delete_list)
        if delete_me:
            for name in delete_list:
                self.logger.debug(f"Deleted {name}")

    def clean(self) -> bool:
        """
        Remove a series of backup files based on the following rules:
        * Keeps all the backups from the last 7 days
        * Keeps the most recent backup from each week of the last month
        * Keeps the most recent backup from each month of the last year
        * Keeps the most recent backup from each year of the remaining years
        """

        # get black and white list
        cleaning = SqlBackupDates(self.get_timestamps())
        white_list = cleaning.white_list
        black_list = cleaning.black_list
        if not black_list:
            self.logger.debug("==> No backup to be deleted.")
            return True

        self.logger.debug(f"==> {len(white_list)} backups will be kept:")
        for timestamp in white_list:
            date_formatted = arrow.Arrow.fromtimestamp(timestamp).humanize()
            self.logger.debug(f"ID: {timestamp} (from {date_formatted})")
            if self.config.debug.sql:
                for f in self.by_timestamp(str(timestamp)):
                    self.logger.debug(f)

        delete_list = []
        self.logger.debug(f"==> {len(black_list)} backups will be deleted:")
        for timestamp in black_list:
            date_formatted = arrow.Arrow.fromtimestamp(timestamp).humanize()
            self.logger.debug(f"ID: {timestamp} (from {date_formatted})")
            for f in self.by_timestamp(str(timestamp)):
                if self.config.debug.database:
                    self.logger.debug(f)
                delete_list.append(f)

        self.delete_backups(delete_list)
        return True

    async def run(self) -> t.NoReturn:
        if not self.config.deployed:
            last_backup = self.get_last_backup()
            if self.config.debug.database and self.sure_delete:
                blobs = [b.name for b in self.storage.sql.list()]
                for b in blobs:
                    self.storage.sql.delete(b)
                # clear_resized_images()
                last_backup = None
            if not last_backup:
                await self.save()
            elif last_backup and self.config.debug.database and not self.sure_delete:
                self.logger.info(f"Restoring last backup - {last_backup}")
                await self.restore_backup(last_backup)
            self.clean()
            self.logger.info("Backups complete.")


depends.set(SqlBackup, SqlBackup())



import typing as t

from acb.config import Config
from acb.depends import depends
from pydantic import SecretStr
from ._base import MonitoringBase
from ._base import MonitoringBaseSettings


class MonitoringSettings(MonitoringBaseSettings):
    enabled: bool = False
    dsn: t.Optional[SecretStr] = None
    sample_rate: float = 0.5

    @depends.inject
    def __init__(self, config: Config = depends(), **values: t.Any) -> None:
        super().__init__(**values)
        self.sample_rate = self.sample_rate if config.deployed else 1.0
        self.enabled = config.deployed or not config.debug.production
        self.dsn = config.secrets.sentry_dsn


class Monitoring(MonitoringBase):
    ...


depends.set(Monitoring, Monitoring())



import typing as t

from acb.depends import depends
from sqlalchemy.pool import NullPool
from ._base import SqlBase
from ._base import SqlBaseSettings


class SqlSettings(SqlBaseSettings):
    _driver: str = "mysql+mysqldb"
    _async_driver: str = "mysql+asyncmy"

    def model_post_init(self, __context: t.Any) -> None:
        self.port = 3306
        self.poolclass = NullPool
        self.pool_pre_ping = True
        self.engine_kwargs = dict(
            poolclass=self.poolclass, pool_pre_ping=self.pool_pre_ping
        )


class Sql(SqlBase):
    ...


depends.set(Sql, Sql())



import typing as t

from acb.adapters.logger import Logger
from acb.depends import depends
from httpx import Response as HttpxResponse
from httpx_cache import AsyncClient
from httpx_cache.cache.redis import RedisCache
from ._base import RequestsBase
from ._base import RequestsBaseSettings


class RequestsSettings(RequestsBaseSettings):
    loggers: t.Optional[list[str]] = ["httpx_caching"]


class Requests(RequestsBase):
    cache: t.Optional[RedisCache] = None

    async def get(self, url: str, timeout: int = 5) -> HttpxResponse:
        async with AsyncClient(cache=self.cache) as client:
            return await client.get(url, timeout=timeout)

    async def post(
        self, url: str, data: dict[str, t.Any], timeout: int = 5
    ) -> HttpxResponse:
        async with AsyncClient(cache=self.cache) as client:
            return await client.post(url, data=data, timeout=timeout)

    async def put(
        self, url: str, data: dict[str, t.Any], timeout: int = 5
    ) -> HttpxResponse:
        async with AsyncClient(cache=self.cache) as client:
            return await client.put(url, data=data, timeout=timeout)

    async def delete(self, url: str, timeout: int = 5) -> HttpxResponse:
        async with AsyncClient(cache=self.cache) as client:
            return await client.delete(url, timeout=timeout)

    @depends.inject
    async def init(self, logger: Logger = depends()) -> None:  # type: ignore
        self.cache = RedisCache(
            redis_url=f"redis://{self.config.cache.host.get_secret_value()}:{self.config.cache.port}/"
            f"{self.config.requests.cache_db}"
        )
        logger.info("Requests adapter loaded")


depends.set(Requests, Requests())



import asyncio
import secrets
import typing as t
from abc import ABC
from abc import abstractmethod
from contextvars import ContextVar
from importlib import import_module
from inspect import currentframe
from pathlib import Path
from secrets import token_bytes
from secrets import token_urlsafe

import nest_asyncio
from acb.actions.encode import dump
from acb.actions.encode import load
from acb.depends import depends
from aiopath import AsyncPath
from async_lru import alru_cache
from inflection import camelize
from inflection import titleize
from inflection import underscore
from pydantic import BaseModel
from pydantic import SecretStr
from pydantic._internal._utils import deep_update
from pydantic.fields import FieldInfo
from pydantic_settings import BaseSettings
from pydantic_settings import SettingsConfigDict
from pydantic_settings.sources import SettingsError

nest_asyncio.apply()

project: str = ""
app_name: str = ""
_deployed: bool = True if Path.cwd().name == "app" else False  # or "srv"?
_pkg_path: AsyncPath = AsyncPath(__file__).parent
_base_path: AsyncPath = AsyncPath.cwd()
_tmp_path: AsyncPath = _base_path / "tmp"
_secrets_path: AsyncPath = _tmp_path / "secrets"
_settings_path: AsyncPath = _base_path / "settings"
_app_settings_path: AsyncPath = _settings_path / "app.yml"
_adapter_settings_path: AsyncPath = _settings_path / "adapters.yml"
_adapters_path = _pkg_path / "adapters"
required_adapters: ContextVar[dict[str, str]] = ContextVar(
    "required_adapters", default={"secrets": "secret_manager", "logger": "loguru"}
)
_app_secrets: ContextVar[set[str]] = ContextVar("_app_secrets", default=set())
loaded_adapters: ContextVar[set[str]] = ContextVar("loaded_adapters", default=set())
available_adapters: ContextVar[dict[str, str | None]] = ContextVar(
    "available_adapters", default={}
)
available_modules: ContextVar[dict[str, dict[str, AsyncPath]]] = ContextVar(
    "available_modules", default={}
)
enabled_adapters: ContextVar[dict[str, str]] = ContextVar(
    "enabled_adapters", default={}
)
package_registry: ContextVar[dict[str, str]] = ContextVar(
    "package_registry", default={}
)
logger_registry: ContextVar[set[str]] = ContextVar("logger_registry", default=set())


async def init_app() -> None:
    for path in (
        _tmp_path,
        _secrets_path,
        _settings_path,
    ):
        await path.mkdir(exist_ok=True)
    enabled_adapters.get().update(required_adapters.get())


asyncio.run(init_app())


def gen_password(size: int = 10) -> str:
    return secrets.token_urlsafe(size)


async def update_available_modules(
    adapters_path: AsyncPath,
) -> dict[str, dict[str, AsyncPath]]:
    _available_modules = available_modules.get()
    for adapter, module in {
        a.stem: {m.stem: m async for m in a.rglob("*.py") if not m.name.startswith("_")}
        async for a in adapters_path.iterdir()
        if await a.is_dir() and not a.name.startswith("__")
    }.items():
        _available_modules.update({adapter: module})
    available_modules.set(_available_modules)
    return _available_modules


async def update_available_adapters(adapters_path: AsyncPath) -> dict[str, str | None]:
    await update_available_modules(adapters_path)
    _available_modules = available_modules.get()
    if not await _adapter_settings_path.exists():
        await dump.yaml(
            {cat: None for cat in _available_modules} | required_adapters.get(),
            _adapter_settings_path,
        )
    _available_adapters = await load.yaml(_adapter_settings_path)
    _available_adapters.update(
        {a: None for a in _available_modules if a not in _available_adapters}
    )
    available_adapters.set(_available_adapters)
    await dump.yaml(_available_adapters, _adapter_settings_path)
    return _available_adapters


async def update_enabled_adapters() -> dict[str, str]:
    _enabled_adapters = enabled_adapters.get()
    _enabled_adapters.update(
        ({a: m for (a, m) in available_adapters.get().items() if m})
    )
    enabled_adapters.set(_enabled_adapters)
    return _enabled_adapters


async def register_package() -> tuple[str, AsyncPath, AsyncPath, dict[str, str | None]]:
    _packages = package_registry.get()
    _pkg_path = AsyncPath(currentframe().f_code.co_filename).parent
    _pkg_name = _pkg_path.name
    _adapters_path = _pkg_path / "adapters"
    package_registry.set({_pkg_name: str(_adapters_path)} | _packages)
    _available_adapters = await update_available_adapters(_adapters_path)
    return _pkg_name, _pkg_path, _adapters_path, _available_adapters


def import_adapter(adapter_name: t.Optional[str] = None) -> t.Any:
    # with suppress(KeyError):
    _pkg_path = Path(currentframe().f_code.co_filename).parent
    if adapter_name is None:
        adapter_name = Path(currentframe().f_back.f_code.co_filename).parent.stem
    _adapter_path = _pkg_path / "adapters" / adapter_name
    _adapter_class_name = camelize(adapter_name)
    _adapter_settings_class_name = f"{_adapter_class_name}Settings"
    _module_name = (
        enabled_adapters.get().get(adapter_name)
        or required_adapters.get().get(adapter_name)
        or ""
    )
    _module_path = _pkg_path / "adapters" / adapter_name / _module_name
    _imported_module = import_module(".".join(_module_path.parts[-4:]))
    _adapter_settings_class = getattr(_imported_module, _adapter_settings_class_name)
    _adapter_class = getattr(_imported_module, _adapter_class_name)
    loaded_adapters.get().add(adapter_name)
    return _adapter_class, _adapter_settings_class
    # if adapter_name in required_adapters.get():
    #     raise SystemExit(f"Required adapter {adapter_name!r} not found.")
    # warn(f"Adapter {adapter_name!r} not found.")


class PydanticBaseSettingsSource(ABC):
    adapter_name: str = "app"

    def __init__(
        self,
        settings_cls: type["Settings"],
        secrets_dir: AsyncPath = _secrets_path,
    ) -> None:
        self.secrets_dir = secrets_dir
        self.settings_cls = settings_cls
        self.config = settings_cls.model_config

    def get_model_secrets(self) -> dict[str, FieldInfo]:
        return {
            "_".join((self.adapter_name, n)): v
            for n, v in self.settings_cls.model_fields.items()
            if v.annotation is SecretStr
        }

    @abstractmethod
    async def __call__(self) -> dict[str, t.Any]:
        self.adapter_name = underscore(
            self.settings_cls.__name__.replace("Settings", "")
        )
        return {}


class InitSettingsSource(PydanticBaseSettingsSource):
    def __init__(
        self, settings_cls: type["Settings"], init_kwargs: dict[str, t.Any]
    ) -> None:
        self.init_kwargs = init_kwargs
        super().__init__(settings_cls)

    async def __call__(self) -> dict[str, t.Any]:
        return self.init_kwargs

    def __repr__(self) -> str:
        return f"InitSettingsSource(init_kwargs={self.init_kwargs!r})"


class FileSecretsSource(PydanticBaseSettingsSource):
    async def get_field_value(self, field_name: str) -> SecretStr | None:
        path = self.secrets_path / field_name
        if await path.is_file():
            return SecretStr((await path.read_text()).strip())
        return None

    async def __call__(self) -> dict[str, t.Any]:
        data = await super().__call__()
        if self.adapter_name == "debug":
            return data
        self.secrets_path = await AsyncPath(self.secrets_dir).expanduser()
        if not await self.secrets_path.exists():
            await self.secrets_path.mkdir(parents=True, exist_ok=True)
        model_secrets = self.get_model_secrets()
        for field_key in model_secrets:
            field_key = field_key.removeprefix(f"{app_name}_")
            if field_key in _app_secrets.get():
                continue
            try:
                field_value = await self.get_field_value(field_key)
            except Exception as e:
                raise SettingsError(
                    f'error getting value for field "{field_key}" from '
                    f'source "{self.__class__.__name__}"'
                ) from e
            if field_value is not None:
                data[field_key] = field_value
                _app_secrets.get().add(field_key)
        return data

    def __repr__(self) -> str:
        return f"FileSecretsSource(_secrets_dir={self.secrets_dir!r})"


class ManagerSecretsSource(PydanticBaseSettingsSource):
    @alru_cache(maxsize=1)
    async def get_manager(self):
        manager = import_adapter("secrets")[0]()
        await manager.init()
        return manager

    async def load_secrets(self) -> t.Any:
        data = {}
        model_secrets = self.get_model_secrets()
        missing_secrets = {
            n: v for n, v in model_secrets.items() if n not in _app_secrets.get()
        }
        if missing_secrets:
            manager = await self.get_manager()
            manager_secrets = await manager.list(self.adapter_name)
            for field_key, field_value in missing_secrets.items():
                stored_field_key = "_".join((app_name, field_key))
                secret_path = _secrets_path / field_key
                if not await secret_path.exists():
                    if field_key not in manager_secrets:
                        await manager.create(
                            stored_field_key, field_value.default.get_secret_value()
                        )
                    secret = await manager.get(stored_field_key)
                    await secret_path.write_text(secret)
                    data[field_key] = secret
                    _app_secrets.get().add(field_key)
        return data

    async def __call__(self) -> t.Any:
        await super().__call__()
        if self.adapter_name == "debug":
            return {}
        data = await self.load_secrets()
        return data


class YamlSettingsSource(PydanticBaseSettingsSource):
    async def load_yml_settings(self) -> t.Any:
        global project, app_name
        if self.adapter_name == "secrets":
            return {}
        yml_path = _settings_path / f"{self.adapter_name}.yml"
        if not await yml_path.exists() and not _deployed:
            dump_settings = {
                name: info.default
                for name, info in self.settings_cls.model_fields.items()
                if (info.annotation is not SecretStr)
                and ("Optional" not in (str(info.annotation)))
            }
            await dump.yaml(dump_settings, yml_path)
        yml_settings = await load.yaml(yml_path)
        if self.adapter_name == "debug":
            for adptr in [
                adptr
                for adptr in enabled_adapters.get()
                if adptr not in yml_settings.keys()
            ]:
                yml_settings[adptr] = False
        if not _deployed:
            await dump.yaml(yml_settings, yml_path)
        if self.adapter_name == "app":
            project = yml_settings["project"]
            app_name = yml_settings["name"]
        return yml_settings

    async def __call__(self) -> dict[str, t.Any]:
        data = await super().__call__()
        if Path.cwd().name != "acb":
            for field_name, field in (await self.load_yml_settings()).items():
                if field is not None:
                    data[field_name] = field
            return data
        return data


class Settings(BaseModel):
    model_config: SettingsConfigDict = SettingsConfigDict(  # type: ignore
        extra="allow",
        arbitrary_types_allowed=True,
        validate_default=True,
        secrets_dir=_secrets_path,
        protected_namespaces=("model_", "settings_"),
    )

    def __getattr__(self, item: str) -> t.Any:
        return super().__getattr__(item)  # type: ignore

    def __init__(
        __pydantic_self__,  # type: ignore
        **values: t.Any,
    ) -> None:
        build_settings = __pydantic_self__._settings_build_values(
            values,
        )
        build_settings = asyncio.run(build_settings)
        super().__init__(**build_settings)

    async def _settings_build_values(
        self,
        init_kwargs: dict[str, t.Any],
        secrets_dir: AsyncPath = _secrets_path,
    ) -> t.Any:
        init_settings = InitSettingsSource(self.__class__, init_kwargs=init_kwargs)
        file_secrets_settings = FileSecretsSource(
            self.__class__,
            secrets_dir=secrets_dir,
        )
        manager_secrets_settings = ManagerSecretsSource(
            self.__class__,
            secrets_dir=secrets_dir,
        )
        yaml_settings = YamlSettingsSource(self.__class__)
        sources = self.settings_customize_sources(
            self.__class__,
            init_settings=init_settings,
            yaml_settings=yaml_settings,
            file_secrets_settings=file_secrets_settings,
            manager_secrets_settings=manager_secrets_settings,
        )
        if sources:
            return deep_update(*reversed([await source() for source in sources]))
        return {}

    @classmethod
    def settings_customize_sources(
        cls,
        settings_cls: t.Type["Settings"],
        init_settings: PydanticBaseSettingsSource,
        yaml_settings: PydanticBaseSettingsSource,
        file_secrets_settings: PydanticBaseSettingsSource,
        manager_secrets_settings: ManagerSecretsSource,
    ) -> tuple[
        PydanticBaseSettingsSource,
        PydanticBaseSettingsSource,
        PydanticBaseSettingsSource,
        ManagerSecretsSource,
    ]:
        return (
            init_settings,
            yaml_settings,
            file_secrets_settings,
            manager_secrets_settings,
        )


class DebugSettings(Settings):
    production: bool = False
    main: bool = False
    secrets: bool = False
    logger: bool = False
    app: bool = False
    adapters: bool = False


class AppSettings(Settings):
    project: str = "test-project"
    name: str = "test"
    title: str = "Test App"
    domain: str = "test_app.local"
    secret_key: SecretStr = SecretStr(token_urlsafe(32))
    secure_salt: SecretStr = SecretStr(str(token_bytes(32)))

    def model_post_init(self, __context: t.Any) -> None:
        self.title = self.title or titleize(self.name)


class Config(BaseSettings, extra="allow"):
    deployed: bool = _deployed
    debug: t.Optional[Settings] = None
    app: t.Optional[Settings] = None

    def __getattr__(self, item: str) -> t.Any:
        return super().__getattr__(item)  # type: ignore

    async def init(self) -> t.Any:
        (
            _pkg_name,
            _pkg_path,
            _adapters_path,
            _available_adapters,
        ) = await register_package()
        _enabled_adapters = await update_enabled_adapters()
        for a in required_adapters.get():
            _enabled_adapters = {a: _enabled_adapters.pop(a)} | _enabled_adapters
        del _enabled_adapters["secrets"]
        enabled_adapters.set(_enabled_adapters)
        self.app = AppSettings()
        self.debug = DebugSettings()


depends.set(Config, Config())



import typing as t

from acb.config import Config
from acb.config import gen_password
from acb.config import Settings
from acb.depends import depends
from pydantic import AnyUrl
from pydantic import field_validator
from pydantic import SecretStr
from pydantic import RedisDsn


class CacheBaseSettings(Settings):
    namespace: t.Optional[str] = None
    db: t.Optional[int] = 1
    host: SecretStr = SecretStr("127.0.0.1")
    password: SecretStr = SecretStr(gen_password())
    _url: t.Optional[AnyUrl | RedisDsn] = None
    default_timeout: int = 86400
    template_timeout: int = 300
    media_timeout: int = 15_768_000
    media_control: str = f"max-age={media_timeout} public"
    port: t.Optional[int] = 6379
    health_check_interval: int = 15
    loggers: t.Optional[list[str]] = []

    @depends.inject
    def __init__(self, config: Config = depends(), **values: t.Any) -> None:
        super().__init__(**values)
        self.namespace = self.namespace or config.app.name or ""
        self.host = SecretStr("127.0.0.1") if not config.deployed else self.host
        self.password = SecretStr("") if not config.deployed else self.password
        self.template_timeout = self.template_timeout if config.deployed else 1

    @field_validator("db")
    def db_less_than_three(cls, v: int) -> int:
        if v < 3 and v != 1:
            raise ValueError("must be greater than 2 (0-2 are reserved)")
        return 1



import typing as t

from acb.depends import depends
from adlfs import AzureBlobFileSystem
from pydantic import SecretStr
from ._base import StorageBase
from ._base import StorageBaseSettings


class StorageSettings(StorageBaseSettings):
    connection_string: SecretStr


class Storage(StorageBase):
    client: t.Any = AzureBlobFileSystem


depends.set(Storage, Storage())



import logging
import sys
import typing as t

from acb.config import Config
from acb.config import enabled_adapters
from acb.config import logger_registry
from acb.depends import depends
from loguru._logger import Core as _Core
from loguru._logger import Logger as _Logger
from ._base import LoggerBaseSettings


class LoggerSettings(LoggerBaseSettings):
    deployed_level: str = "ERROR"
    serialize: t.Optional[bool] = False
    format: t.Optional[dict[str, str]] = dict(
        time="<b><e>[</e> <w>{time:YYYY-MM-DD HH:mm:ss.SSS}</w> <e>]</e></b>",
        level=" <level>{level:>8}</level>",
        sep=" <b><w>in</w></b> ",
        name="{name:>28}",
        line="<b><e>[</e><w>{line:^5}</w><e>]</e></b>",
        message="  <level>{message}</level>",
    )
    level_per_module: t.Optional[dict[str, str]] = {}
    settings: t.Optional[dict[str, t.Any]] = {}

    @depends.inject
    def __init__(self, config: Config = depends(), **values: t.Any) -> None:
        super().__init__(**values)
        self.level_per_module = {
            m: "DEBUG" if v is True else "INFO"
            for (m, v) in config.debug.model_dump().items()
        }
        self.settings = dict(
            filter=self.level_per_module,
            format="".join(self.format.values()),
            enqueue=True,
            backtrace=True,
            serialize=self.serialize,
        )


class Logger(_Logger):
    def __init__(self) -> None:
        super().__init__(
            core=_Core(),
            exception=None,
            depth=0,
            record=False,
            lazy=False,
            colors=False,
            raw=False,
            capture=True,
            patchers=[],
            extra={},
        )

    @depends.inject
    async def init(self, config: Config = depends()) -> None:
        def patching(record) -> None:  # type: ignore
            record["extra"]["mod_name"] = ".".join(record["name"].split(".")[0:-1])

        self.remove()
        # self.patch(patching)
        self.add(sys.stderr, **config.logger.settings)
        if config.deployed:
            self.level = config.logger.deployed_level
        self.info("Logger adapter loaded")
        if config.debug.logger:
            self.debug("debug")
            self.info("info")
            self.warning("warning")
            self.error("error")
            self.critical("critical")


depends.set(Logger, Logger())


class InterceptHandler(logging.Handler):
    def __init__(self, logger_name: str) -> None:
        super().__init__()
        self.logger_name = logger_name

    @depends.inject
    def emit(
        self, record: logging.LogRecord, logger: Logger = depends()  # type: ignore
    ) -> None:
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno
        frame, depth = logging.currentframe(), 2
        while frame.f_code.co_filename == logging.__file__:
            frame = frame.f_back
            depth += 1
            enabled_logger = enabled_adapters.get()["logger"]
            if enabled_logger == "loguru":
                logger.patch(
                    lambda record: record.update(name=self.logger_name)  # type: ignore
                ).opt(
                    depth=depth,
                    exception=record.exc_info,
                ).log(
                    level, record.getMessage()
                )
            # elif enabled_logger == "structlog":
            #     ...


def register_loggers(loggers: list[str]) -> None:
    for logger in loggers:
        _logger = logging.getLogger(logger)
        _logger.handlers.clear()
        _logger.handlers = [InterceptHandler(_logger.name)]
    return logger_registry.get().update(loggers)



# from redis_om import get_redis_connection
# from redis_om import Migrator

from ._base import NosqlBaseSettings
from ._base import NosqlBase
from pydantic import field_validator


class NosqlSettings(NosqlBaseSettings):
    cache_db: int = 0

    @field_validator("cache_db")
    def cache_db_not_zero(cls, v: int) -> int:
        if v < 3 and v != 0:
            raise ValueError("must be > 3 (0-2 are reserved)")
        return 0


class Nosql(NosqlBase):
    def __init__(self) -> None:
        ...
        # Migrator().run()



import asyncio
import typing as t

import asyncio_atexit
from acb.actions.encode import dump
from acb.actions.encode import load
from acb.adapters.logger import Logger
from acb.config import Config
from acb.debug import debug
from acb.depends import depends
from cashews.serialize import register_type
from cashews.wrapper import Cache as CashewsCache
from pydantic import RedisDsn
from ._base import CacheBaseSettings


class CacheSettings(CacheBaseSettings):
    def model_post_init(self, __context: t.Any) -> None:
        self._url = RedisDsn(
            f"redis://{self.host.get_secret_value()}:{self.port}/{self.db}"
        )


class Cache(CashewsCache):
    config: Config = depends()
    logger: Logger = depends()  # type: ignore

    async def encoder(self, value: t.Any, *args: object, **kwargs: object) -> bytes:
        return await dump.msgpack(
            value,
            secret_key=self.config.app.secret_key,
            secure_salt=self.config.app.secure_salt,
        )

    async def decoder(self, value: bytes, *args: object, **kwargs: object) -> t.Any:
        return await load.msgpack(
            value,
            secret_key=self.config.app.secret_key,
            secure_salt=self.config.app.secure_salt,
        )

    async def init(self, *args: object, **kwargs: object) -> t.NoReturn:
        await super().init(
            str(self.config.cache._url),
            password=self.config.cache.password.get_secret_value(),
            client_side=True,
            client_side_prefix=f"{self.config.app.name}:",
        )
        register_type(t.Any, self.encoder, self.decoder)

        async def close_cache_session() -> None:
            self.logger.debug("Closing cache session...")
            loop = asyncio.get_running_loop()
            debug(loop.is_running())
            await self.close()
            self.logger.debug("Cache session closed")

        asyncio_atexit.register(close_cache_session)

        self.logger.info("Cache adapter loaded")


depends.set(Cache, Cache())



import typing as t

from bevy import dependency
from bevy import get_repository
from bevy import inject as inject_dependency


class Depends:
    @staticmethod
    def inject(func: t.Callable[..., t.Any]) -> t.Callable[..., t.Any]:
        return inject_dependency(func)

    @staticmethod
    def set(class_: t.Any, value: t.Any) -> t.Any:
        return get_repository().set(class_, value)

    @staticmethod
    def get(class_: t.Any) -> t.Any:
        return get_repository().get(class_)

    def __call__(self, *args: t.Any, **kwargs: t.Any):
        return dependency()


depends = Depends()



import typing as t

from sqlmodel import Field

# from sqlmodel import select
from sqlmodel import SQLModel

from acb.depends import depends
from inflection import underscore
from sqlalchemy.ext.declarative import declared_attr
from ulid import ULID
from ._base import SqlBase


class SqlModel(SQLModel):
    __table_args__ = {"extend_existing": True}
    __mapper_args__ = {"always_refresh": True}
    id: t.Optional[ULID] = Field(default_factory=ULID, primary_key=True)

    # @property  # hybrid
    # def date_created(self) -> arrow.Arrow:
    #     return arrow.get(ULID(self.id).timestamp)

    @declared_attr
    def __tablename__(cls) -> t.Any:  # type: ignore
        return underscore(cls.__name__)

    class Config:
        arbitrary_types_allowed = True
        extra = "allow"

    @depends.inject
    async def save(self, sql: SqlBase = depends()) -> None:  # type: ignore
        async with sql.get_session() as session:
            session.add(self)
            await session.commit()



import typing as t
from contextlib import asynccontextmanager
from functools import cached_property

from sqlmodel import SQLModel
from sqlmodel.ext.asyncio.session import AsyncSession

from acb.adapters.logger import Logger
from acb.config import Config
from acb.config import gen_password
from acb.config import Settings
from acb.debug import debug
from acb.depends import depends
from aioconsole import ainput
from aioconsole import aprint
from pydantic import SecretStr
from sqlalchemy import inspect
from sqlalchemy.engine import URL
from sqlalchemy.exc import OperationalError
from sqlalchemy.ext.asyncio import AsyncConnection
from sqlalchemy.ext.asyncio import AsyncEngine
from sqlalchemy.ext.asyncio import create_async_engine
from sqlalchemy_utils import create_database
from sqlalchemy_utils import database_exists
from sqlalchemy_utils import drop_database


class SqlBaseSettings(Settings):
    _driver: str
    _async_driver: str
    port: t.Optional[int] = 3306
    pool_pre_ping: t.Optional[bool] = False
    poolclass: t.Optional[t.Any] = None
    host: SecretStr = SecretStr("127.0.0.1")
    user: SecretStr = SecretStr("root")
    password: SecretStr = SecretStr(gen_password())
    _url: t.Optional[URL] = None
    _async_url: t.Optional[URL] = None
    engine_kwargs: t.Optional[dict[str, t.Any]] = {}
    backup_bucket: str | None = None
    loggers: t.Optional[list[str]] = [
        "sqlalchemy.engine",
        "sqlalchemy.orm",
        "sqlalchemy.pool",
        "sqlalchemy.dialects",
    ]

    @depends.inject
    def __init__(self, config: Config = depends(), **values: t.Any) -> None:
        super().__init__(**values)
        _url_kwargs = dict(
            drivername=self._driver,
            username=self.user.get_secret_value(),
            password=self.password.get_secret_value(),
            host=("127.0.0.1" if not config.deployed else self.host.get_secret_value()),
            port=self.port,
            database=config.app.name,
        )
        self._url = URL.create(**_url_kwargs)  # type: ignore
        _async_url_kwargs = dict(drivername=self._async_driver)
        self._async_url = URL.create(**(_url_kwargs | _async_url_kwargs))  # type:ignore


class SqlBase:
    config: Config = depends()
    logger: Logger = depends()  # type: ignore

    @cached_property
    def engine(self) -> AsyncEngine:
        return create_async_engine(
            self.config.sql._async_url, **self.config.sql.engine_kwargs
        )

    @cached_property
    def session(self) -> AsyncSession:
        return AsyncSession(self.engine, expire_on_commit=False)

    @depends.inject
    async def create(self, demo: bool = False) -> None:
        try:
            exists = database_exists(self.config.sql._url)
            if exists:
                self.logger.debug("Database exists")
        except OperationalError:
            exists = False

        if (
            self.config.debug.sql
            and not (self.config.deployed or self.config.debug.production)
            and exists
        ):
            msg = (
                "\n\nRESETTING THE DATABASE WILL CAUSE ALL OF YOUR"
                " CURRENT DATA TO BE LOST!\n"
            )
            if demo:
                msg = (
                    "\nBy running this module,\n\nYOUR DATABASE WILL BE DELETED AND "
                    "REPLACED WITH DEMO DATA.\nALL OF YOUR CURRENT DATA WILL BE LOST!\n"
                )
            await aprint(msg)
            delete_db = await ainput("Would you like to reset the database? (y/N) ")
            await aprint()
            if delete_db.upper().strip() == "Y":
                drop_database(self.config.sql._url)
                self.logger.warning("Database dropped")
                exists = database_exists(self.config.sql._url)
        if not exists:
            create_database(self.config.sql._url)
            self.logger.debug("Database created")

    @asynccontextmanager
    async def get_session(self) -> t.AsyncGenerator[AsyncSession, None]:
        async with self.session as sess:
            yield sess

    @asynccontextmanager
    async def get_conn(self) -> t.AsyncGenerator[AsyncConnection, None]:
        async with self.engine.begin() as conn:
            yield conn

    @staticmethod
    def get_table_names(conn: object) -> list[str]:
        inspector = inspect(conn)
        return inspector.get_table_names() or []

    async def init(
        self,
        demo: bool = False,
    ) -> None:
        self.logger.info("Loading database adapter...")
        await self.create(demo)
        async with self.get_conn() as conn:
            # if self.config.debug.sql:
            #     sql = text("DROP TABLE IF EXISTS alembic_version")
            #     await conn.execute(sql)
            self.logger.debug("Creating database tables...")
            await conn.run_sync(SQLModel.metadata.create_all)
            if self.config.debug.sql:
                table_names = await conn.run_sync(self.get_table_names)
                debug(table_names)
        self.logger.info("Database adapter loaded")



import typing as t

from acb.adapters.logger import Logger
from acb.config import Config
from acb.config import Settings
from acb.depends import depends
from aiopath import AsyncPath
from google.cloud.exceptions import NotFound


# CORS policy for upload bucket - upload-cors.json
#
# [
#     {
#       "origin": ["*"],
#       "method": ["*"],
#       "responseHeader": ["*"],
#       "maxAgeSeconds": 600
#     }
# ]


class StorageBaseSettings(Settings):
    prefix: t.Optional[str] = None
    user_project: t.Optional[str] = None  # used for billing
    buckets: dict[str, str] = {}
    loggers: t.Optional[list[str]] = []

    @depends.inject
    def __init__(self, config: Config = depends(), **values: t.Any) -> None:
        super().__init__(**values)
        self.prefix = self.prefix or config.app.name or ""
        self.user_project = (
            self.user_project or config.app.name or ""
        )  # used for billing


class StorageBucket:
    config: Config = depends()
    logger: Logger = depends()  # type: ignore

    def __init__(
        self,
        client: t.Any,
        bucket: str,
        prefix: t.Optional[str] = None,
    ) -> None:
        self.client = client
        self.bucket = self.config.storage.buckets[bucket]
        self.prefix = prefix or self.config.storage.prefix
        self.root = AsyncPath(f"{self.bucket}/{self.prefix}")

    def get_name(self, path: AsyncPath) -> str:
        return path.name

    def get_path(self, path: AsyncPath) -> str:
        return str(self.root / path)

    def get_url(self, path: AsyncPath):
        return self.client.url(self.get_path(path))

    async def get_date_created(self, path: AsyncPath) -> t.Any:
        return (await self.stat(path))["timeCreated"]

    async def get_date_updated(self, path: AsyncPath) -> t.Any:
        return (await self.stat(path))["updated"]

    async def get_size(self, path: AsyncPath) -> int:
        return (await self.stat(path))["size"]

    async def get_signed_url(self, path: AsyncPath, expires: int = 3600) -> t.Any:
        return await self.client._sign(self.get_path(path), expires=expires)

    async def stat(self, path: AsyncPath) -> t.Any:
        return await self.client._info(self.get_path(path))

    async def list(self, dir_path: AsyncPath) -> t.Any:
        return await self.client._ls(self.get_path(dir_path))

    async def exists(self, path: AsyncPath) -> t.Any:
        return await self.client._exists(self.get_path(path))

    async def create_bucket(self, path: AsyncPath) -> t.Any:
        return await self.client._mkdir(
            self.get_path(path),
            create_parents=True,
            enable_versioning=False,
            # acl="public-read",
        )

    async def read(self, path: AsyncPath) -> t.Any:
        stor_path = self.get_path(path)
        self.logger.debug(f"Getting {stor_path}...")
        try:
            data = await self.client._cat_file(stor_path)
        except (NotFound, FileNotFoundError):
            raise FileNotFoundError  # for jinja loaders
        self.logger.debug(f"Got - {stor_path}")
        return data

    async def write(self, path: AsyncPath, data: t.Any) -> t.Any:
        stor_path = self.get_path(path)
        self.logger.debug(f"Saving {stor_path}...")
        await self.client._pipe_file(stor_path, data)
        self.logger.debug(f"Saved - {stor_path}")

    async def delete(self, path: AsyncPath) -> t.Any:
        stor_path = self.get_path(path)
        self.logger.debug(f"Deleting {stor_path}...")
        await self.client._rm_file(stor_path)
        self.logger.debug(f"Deleted - {stor_path}")

    # async def get_name(self, name: str) -> t.Any:
    #     ...
    #
    # async def get_path(self, name: str) -> t.Any:
    #     ...
    #
    # def get_size(self, name: str) -> int:
    #     ...
    #
    # def open(self, name: str) -> t.BinaryIO:
    #     ...
    #
    # def write(self, file: t.BinaryIO, name: str) -> str:
    #     ...


class StorageBase:
    client: t.Any
    config: Config = depends()
    logger: Logger = depends()  # type: ignore

    async def init(self) -> None:
        self.client = self.client(asynchronous=True)
        for bucket in self.config.storage.buckets:
            setattr(self, bucket, StorageBucket(self.client, bucket))
            self.logger.debug(f"{bucket.title()} storage bucket initialized")
        self.logger.info("Storage adapter loaded")


class StorageFile:
    def __init__(self, *, name: str, storage: StorageBucket) -> None:
        self._storage: StorageBucket = storage
        self._name: str = name

    @property
    def name(self) -> str:
        return self._storage.get_name(AsyncPath(self._name))

    @property
    def path(self) -> str:
        return self._storage.get_path(AsyncPath(self._name))

    @property
    async def size(self) -> t.Any:
        return await self._storage.get_size(AsyncPath(self._name))

    # async def open(self) -> t.Any:
    #     return await self._storage.open(self._name)

    async def write(self, file: t.BinaryIO) -> t.Any:
        return await self._storage.write(path=AsyncPath(self._name), data=file)

    def __str__(self) -> str:
        return self.path


class StorageImage(StorageFile):
    def __init__(
        self, *, name: str, storage: StorageBucket, height: int, width: int
    ) -> None:
        super().__init__(name=name, storage=storage)
        self._width = width
        self._height = height

    @property
    def height(self) -> int:
        return self._height

    @property
    def width(self) -> int:
        return self._width



import asyncio

from acb.config import Config
from acb.config import enabled_adapters
from acb.config import loaded_adapters
from acb.config import package_registry
from acb.depends import depends
from inflection import camelize
from pluginbase import PluginBase

adapters_base = PluginBase(package="acb.adapters")
adapters_source = adapters_base.make_plugin_source(
    searchpath=[], identifier="acb-adapters"
)


@depends.inject
async def load_adapter(adapter_name: str, config: Config = depends()) -> None:
    adapter_class_name = camelize(adapter_name)
    adapters_source.searchpath = list(package_registry.get().values())
    # ic("loading adapter", adapter_name)
    # ic(adapters_source.searchpath)
    with adapters_source:
        adapter_module = adapters_source.load_plugin(adapter_name)
    # ic(adapter_module)
    # required = getattr(adapter_module, "requires", [])
    # for required_adapter in required:
    #     if required_adapter not in loaded_adapters.get():
    #         await load_adapter(required_adapter)
    #         required_adapters.get().update(required_adapter)
    adapter_class = getattr(adapter_module, adapter_class_name)
    adapter_settings = getattr(adapter_module, f"{adapter_class_name}Settings")
    setattr(config, adapter_name, adapter_settings())
    initialized_adapter = depends.get(adapter_class)
    await initialized_adapter.init()


@depends.inject
async def main(config: Config = depends()) -> None:
    await config.init()
    for adapter_name in enabled_adapters.get():
        if adapter_name in loaded_adapters.get():
            continue
        await load_adapter(adapter_name)
    # ic(available_modules.get())
    # ic(available_adapters.get())
    # ic(required_adapters.get())
    # ic(enabled_adapters.get())
    # ic(package_registry.get())


loop = asyncio.new_event_loop() or asyncio.get_running_loop()
loop.run_until_complete(main())  # type: ignore



from asyncio import sleep
from contextlib import suppress
from warnings import catch_warnings
from warnings import filterwarnings

from acb.depends import depends
from google.api_core.exceptions import BadRequest
from google.api_core.exceptions import Conflict
from google.cloud.dns import Changes
from google.cloud.dns import Client as DnsClient
from validators import domain
from validators.utils import ValidationError
from ._base import DnsBase
from ._base import DnsBaseSettings
from ._base import DnsRecord


class DnsSettings(DnsBaseSettings):
    ...


class Dns(DnsBase):
    async def init(self) -> None:
        with catch_warnings():
            filterwarnings("ignore", category=Warning)
            self.client = DnsClient(project=self.config.app.project)
        self.logger.info("Dns adapter loaded")

    async def create_zone(self) -> None:
        self.zone = self.client.zone(self.config.app.name, f"{self.config.app.domain}.")
        if not self.zone.exists():
            self.logger.info(f"Creating gdns zone '{self.config.app.name}...")
            self.zone.create()
            self.logger.info(f"Zone '{self.zone.name}' successfully created")
        else:
            self.logger.info(f"Zone for '{self.zone.name}' exists")

    def list_records(self) -> list[DnsRecord]:
        records = self.zone.list_resource_record_sets()
        records = [
            DnsRecord.model_validate(
                dict(
                    name=record.name,
                    type=record.record_type,
                    ttl=record.ttl,
                    rrdata=record.rrdatas,
                )
            )
            for record in records
        ]
        return records

    async def wait_for_changes(self, changes: Changes) -> None:
        while changes.status != "done":
            await sleep(3)
            changes.reload()  # API request

    async def apply_changes(self, changes: Changes):
        try:
            changes.create()  # API request
            await self.wait_for_changes(changes)
        except (Conflict, BadRequest) as err:
            change = changes.additions[0]  # type: ignore
            if change.name.split(".")[1] != self.config.app.project:
                raise err
            self.logger.info("Development domain detected - no changes made")

    async def create_records(self, records: list[DnsRecord] | DnsRecord) -> None:
        if isinstance(records, DnsRecord):
            records = [records]
        changes = self.zone.changes()
        current_record_sets = []
        new_record_sets = []
        for record in records:
            if not record.name.endswith("."):
                record.name = f"{record.name}."
            record.rrdata = (
                [record.rrdata]
                if not isinstance(record.rrdata, list)
                else record.rrdata
            )
            for i, r in enumerate(record.rrdata):
                with suppress(ValidationError):
                    if isinstance(r, str) and domain(r) and not r.endswith("."):
                        r = f"{r}."
                        record.rrdata[i] = r
                    if record.type == "TXT":
                        record.rrdata[i] = f'"{r}"'
            current_record = [
                r
                for r in self.list_records()
                if r.name == record.name and r.type == record.type
            ]
            if len(current_record) == 1:
                # print(current_record[0])
                current_record = current_record[0]
                if current_record.__dict__ == record.__dict__:
                    continue
                current_record_set = self.zone.resource_record_set(
                    name=current_record.name,
                    record_type=current_record.type,
                    ttl=current_record.ttl,
                    rrdatas=current_record.rrdata,
                )
                current_record_sets.append(current_record_set)
                self.logger.info(f"Deleting - {current_record}")
            record_set = self.zone.resource_record_set(
                name=record.name,
                record_type=record.type,
                ttl=record.ttl,
                rrdatas=record.rrdata,
            )
            new_record_sets.append(record_set)
            self.logger.info(f"Creating - {record}")
        if current_record_sets:
            for record_set in current_record_sets:
                changes.delete_record_set(record_set)
            self.logger.info("Deleting record sets")
            await self.apply_changes(changes)
        changes = self.zone.changes()
        if new_record_sets:
            for record_set in new_record_sets:
                changes.add_record_set(record_set)
            self.logger.info("Creating record sets")
            await self.apply_changes(changes)
        else:
            self.logger.info("No DNS changes detected")


depends.set(Dns, Dns())



import typing as t

from acb.depends import depends
from gcsfs.core import GCSFileSystem
from ._base import StorageBase
from ._base import StorageBaseSettings


class StorageSettings(StorageBaseSettings):
    ...


class Storage(StorageBase):
    client: t.Any = GCSFileSystem


depends.set(Storage, Storage())



import typing as t
from abc import ABC
from abc import abstractmethod

from acb.config import Config
from acb.config import Settings
from acb.depends import depends
from pydantic import field_validator


class RequestsBaseSettings(Settings):
    cache_db: t.Optional[int] = 2

    @field_validator("cache_db")
    def cache_db_less_than_three(cls, v: int) -> int:
        if v < 3 and v != 2:
            raise ValueError("must be greater than 2 (0-2 are reserved)")
        return 2


class RequestsBase(ABC):
    config: Config = depends()

    @abstractmethod
    async def init(self) -> None:
        raise NotImplementedError

    @abstractmethod
    async def get(self, url: str, timeout: int) -> t.Any:
        raise NotImplementedError

    @abstractmethod
    async def post(self, url: str, data: dict[str, t.Any], timeout: int) -> t.Any:
        raise NotImplementedError

    @abstractmethod
    async def put(self, url: str, data: dict[str, t.Any], timeout: int) -> t.Any:
        raise NotImplementedError

    @abstractmethod
    async def delete(self, url: str, timeout: int) -> t.Any:
        raise NotImplementedError



import typing as t
from abc import ABC
from abc import abstractmethod

from acb.adapters.logger import Logger
from acb.config import Config
from acb.config import Settings
from acb.depends import depends
from pydantic import BaseModel


class DnsRecord(BaseModel):
    name: t.Optional[str] = None
    type: str = "TXT"
    ttl: int = 300
    rrdata: t.Optional[str | list[t.Any]] = None

    @depends.inject
    def __init__(self, config: Config = depends(), **data: t.Any) -> None:
        super().__init__(**data)
        self.name = f"mail.{config.app.domain}"


class DnsBaseSettings(Settings):
    ...


class DnsBase(ABC):
    config: Config = depends()
    logger: Logger = depends()  # type: ignore
    client: t.Optional[t.Any] = None
    zone: t.Optional[t.Any] = None

    @abstractmethod
    async def init(self) -> None:
        raise NotImplementedError

    @abstractmethod
    async def create_zone(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def list_records(self) -> list[DnsRecord]:
        raise NotImplementedError

    @abstractmethod
    async def create_records(self, records: list[DnsRecord] | DnsRecord) -> None:
        raise NotImplementedError



import logging
import typing as t
from pathlib import Path
from pprint import pformat

from acb.adapters.logger import Logger
from acb.config import Config
from acb.config import enabled_adapters
from acb.depends import depends
from icecream import colorizedStderrPrint

config = depends.get(Config)
logger = depends.get(Logger)


def get_calling_module() -> Path | None:
    mod = logging.currentframe().f_back.f_back.f_back.f_code.co_filename
    mod = Path(mod).parent
    debug_mod = getattr(config.debug, mod.stem, None) or (
        mod.stem == config.app.name and config.debug.main
    )
    return mod if debug_mod else None


def patch_record(mod: Path, msg: str) -> None:
    if enabled_adapters.get()["logger"] == "loguru":
        logger.patch(
            lambda record: record.update(name=mod.name),  # type: ignore
        ).debug(msg)


def print_debug_info(msg: str) -> t.Any:
    mod = get_calling_module()
    if mod:
        if config.deployed or config.debug.production:
            patch_record(mod, msg)
        else:
            colorizedStderrPrint(msg)


try:
    from icecream import ic as debug

    debug_args = dict(
        outputFunction=print_debug_info,
        argToStringFunction=lambda o: pformat(o, sort_dicts=False),
    )
    debug.configureOutput(
        prefix="    debug:  ",
        includeContext=True,
        **debug_args,
    )
    if config.deployed or config.debug.production:
        debug.configureOutput(
            prefix="",
            includeContext=False,
            **debug_args,
        )

except ImportError:

    def debug(*a: t.Any) -> t.Any:
        fake_ic = a[0] if len(a) == 1 else a
        return None if not a else fake_ic



import typing as t
from time import perf_counter

from acb.adapters.logger import Logger
from acb.config import Settings
from acb.depends import depends


class MonitoringBaseSettings(Settings):
    ...


class MonitoringBase:
    ...


@depends.inject
def timeit(func: t.Any, logger: Logger = depends()) -> t.Any:  # type: ignore
    def wrapped(*args: t.Any, **kwargs: t.Any) -> t.Any:
        start = perf_counter()
        result = func(*args, **kwargs)
        end = perf_counter()
        logger.debug(f"Function '{func.__name__}' executed in {end - start} s")
        return result

    return wrapped
